// Package graphql provides a GraphQL client implementation.
//
// For more information, see package github.com/shurcooL/githubv4,
// which is a specialized version targeting GitHub GraphQL API v4.
// That package is driving the feature development.
//
// For now, see README for more details.
package graphql // import "github.com/shurcooL/graphql"
